@@ -91,6 +91,8 @@ GetOptions(\%opts,
            'pnp1|A005234',
            'pnm1|A006794',
            'euclid|A018239',
+           'circular|A068652',
+           'panaitopol|A027862',
            'provable',
            'nompugmp',   # turn off MPU::GMP for debugging
            'version',
@@ -157,6 +159,9 @@ if ($start > $end) {
     if (exists $opts{'palindromic'}) {
       $segment_size = 10**length($start) - $start - 1; # all n-digit numbers
     }
+    if (exists $opts{'panaitopol'}) {
+      $segment_size = (~0 == 4294967295) ? 2147483648 : int(10**12);
+    }
 
     my $seg_start = $start;
     my $seg_end = int($start + $segment_size);
@@ -276,6 +281,25 @@ sub cuban_primes {
   @cprimes;
 }
 
+sub panaitopol_primes {
+  my ($start, $end) = @_;
+
+  my @init;
+  push @init,  5 if $start <=  5 && $end >=  5;
+  push @init, 13 if $start <= 13 && $end >= 13;
+  return @init if $end < 41;
+  my $nbeg = ($start <= 41)  ?  4  :  int( sqrt( ($start-1)/2) );
+  my $nend = int( sqrt(($end-1)/2) );
+  $nbeg++ while (2*$nbeg*($nbeg+1)+1) < $start;
+  $nend-- while (2*$nend*($nend+1)+1) > $end;
+  # TODO: BigInts
+  return @init,
+         grep { is_prime($_) }
+         grep { ($_%5) && ($_%13) && ($_%17) && ($_%29) && ($_%37) }
+         map { 2*$_*($_+1)+1 }
+         $nbeg .. $nend;
+}
+
 sub lucky_primes {
   my ($start, $end) = @_;
   # First do a (very basic) lucky number sieve to generate A000959.
@@ -308,7 +332,8 @@ sub lucky_primes {
     } while ($index <= $#lucky);
   }
   # Then restrict to primes to get A031157.
-  grep { $_ >= $start && is_prime($_) } @lucky;
+  shift @lucky while $lucky[0] < $start;
+  grep { is_prime($_) } @lucky;
 }
 
 # This is not a general palindromic digit function!
@@ -332,26 +357,23 @@ sub ndig_palindromes {
 }
 
 # See: http://en.wikipedia.org/wiki/Pillai_prime
-# This is quite slow.
 sub is_pillai {
   my $p = shift;
-  return 0 if $p < 2;
-  # Simpler but much slower:
-  #   foreach my $n (grep { ($p % $_) != 1 } (2 .. $p-1)) {
-  #     return 1 if Math::BigInt->new($n)->bfac()->bmod($p) == ($p-1);
-  #   }
-  # About 20% faster but sucks memory:
-  #   foreach my $n (grep { ($p % $_) != 1 } (2 .. $p-1)) {
-  #     $fac_c[$n] = Math::BigInt->new($n)->bfac() if !defined $fac_c[$n];
-  #     return 1 if ($fac_c[$n] % $p) == ($p-1);
-  #   }
-  # Once p gets large (say 20,000+), then calculating and storing n! is
-  # unreasonable, and the code below will be much faster.
-  my $n_factorial_mod_p = Math::BigInt->bone();
-  for (my $n = 2; $n < $p; $n++) {
-    $n_factorial_mod_p->bmul($n)->bmod($p);
-    next if $p % $n == 1;
-    return 1 if $n_factorial_mod_p == ($p-1);
+  return 0 if $p <= 2;
+  my $half_word = (~0 == 4294967295) ? 65535 : 4294967295;
+  if ($p <= $half_word) {
+    my $nfac = 1;
+    for (my $n = 2; $n < $p; $n++) {
+      $nfac = ($nfac * $n) % $p;
+      return 1 if $nfac == $p-1 && ($p % $n) != 1;
+    }
+  } else {
+    # Must use bigints.  Very slow.
+    my $n_factorial_mod_p = Math::BigInt->bone();
+    for (my $n = Math::BigInt->new(2); $n < $p; $n++) {
+      $n_factorial_mod_p->bmul($n)->bmod($p);
+      return 1 if $n_factorial_mod_p == ($p-1) && ($p % $n) != 1;
+    }
   }
   0;
 }
@@ -370,6 +392,18 @@ sub is_good_prime {
   1;
 }
 
+# Assumes the input is prime.  Returns 1 if all digit rotations are prime.
+sub is_circular_prime {
+  my $p = shift;
+  return 1 if $p < 10;
+  return 0 if $p =~ tr/024568//;
+  # TODO: BigInts
+  foreach my $rot (1 .. length($p)-1) {
+    return 0 unless is_prime( substr($p, $rot) . substr($p, 0, $rot) );
+  }
+  1;
+}
+
 sub merge_primes {
   my ($genref, $pref, $name, @primes) = @_;
   if (!defined $$genref) {
@@ -410,6 +444,9 @@ sub gen_and_filter {
   if (exists $opts{'cuban2'}) {
     merge_primes(\$gen, $p, 'cuban2', cuban_primes($start, $end, 2));
   }
+  if (exists $opts{'panaitopol'}) {
+    merge_primes(\$gen, $p, 'panaitopol', panaitopol_primes($start, $end));
+  }
   if (exists $opts{'palindromic'}) {
     if (!defined $gen) {
       foreach my $d (length($start) .. length($end)) {
@@ -485,6 +522,9 @@ sub gen_and_filter {
   if (exists $opts{'pnp1'}) {
     @$p = grep { is_prime( primorial(Math::BigInt->new($_))+1 ) } @$p;
   }
+  if (exists $opts{'circular'}) {
+    @$p = grep { is_circular_prime($_) } @$p;
+  }
   if (exists $opts{'pillai'}) {
     @$p = grep { is_pillai($_); } @$p;
   }
@@ -511,6 +551,7 @@ sub find_mod210_restriction {
     sexy       => {min=> 7, mod=>[11,13,17,23,31,37,41,47,53,61,67,73,83,97,101,103,107,121,131,137,143,151,157,163,167,173,181,187,191,193]},
     safe       => {min=>11, mod=>[17,23,47,53,59,83,89,107,137,143,149,167,173,179,209]},
     sophie     => {min=> 5, mod=>[11,23,29,41,53,71,83,89,113,131,149,173,179,191,209]},
+    panaitopol => {min=> 5, mod=>[1,11,13,41,43,53,61,71,83,103,113,131,151,173,181,193]},
     # Nothing for good, pillai, palindromic, fib, lucas, mersenne, primorials
   );
 
@@ -579,6 +620,8 @@ to only those primes additionally meeting these conditions:
   --pnp1       Primorial+1      p#+1 is prime
   --pnm1       Primorial-1      p#-1 is prime
   --euclid     Euclid           pn#+1 is prime
+  --circular   Circular         all digit rotations of p are prime
+  --panaitopol Panaitopol       p = (x^4-y^4)/(x^3+y^3) for some x,y
   --provable                    Ensure all primes are provably prime
 
 Note that options can be combined, e.g. display only safe twin primes.
